{- This module acts as a bridge between HarmTrace and Haskore 
   It converts types between the two models and exports 
   midi files using Haskore types -}

module HarmTrace.Harmony.HaskoreBridge  where

--Haskore
import           Haskore.Basic.Scale 
import qualified Haskore.Composition.Drum               as Drum
import qualified Haskore.Composition.Chord              as Chord
import           Haskore.Basic.Dynamics (Velocity)      
import           Haskore.Melody                         as Melody
import           Haskore.Melody.Standard                as MelodyS
import           Haskore.Music.GeneralMIDI              as MidiMusic
import qualified Haskore.Music                          as Music 
import           Haskore.Basic.Pitch (Class(..), Relative, fromInt, Octave)
import           Haskore.Interface.MIDI.Render          as Render
import           Haskore.Interface.MIDI.Read            as ReadMidi
import qualified Haskore.Interface.MIDI.Write           as WriteMidi 
import           Haskore.Interface.MIDI.Note            as Note
import           Haskore.Interface.MIDI.InstrumentMap   as InstrMap
import           Haskore.Basic.Duration                           
import qualified Haskore.Basic.Interval                 as Interval
import qualified Haskore.Performance.Context            as Context
import qualified Haskore.Music.Rhythmic                 as RhyMusic
import qualified Numeric.NonNegative.Wrapper            as NonNeg
import qualified Medium.Controlled.List                 as CtrlMediumList
import qualified Medium.Plain.List                      as PlainMediumList
import           Haskore.Performance                    as Performance 
--MIDI
import qualified Sound.MIDI.General                     as GM
import           Sound.MIDI.File.Load                   as LoadMidi
import qualified Sound.MIDI.File.Save                   as SaveMidi
import           Sound.MIDI.File                        as MidiFile    
import           Sound.MIDI.Message.Channel             as ChannelMsg  
import qualified Sound.MIDI.File.Event.Meta             as MetaEvent
import qualified Sound.MIDI.File.Event                  as Event       
import           Sound.MIDI.KeySignature                as KeySignature
--Misc
import Data.Tuple.HT (fst3, snd3, thd3, )
import qualified Data.Accessor.Basic                    as Accessor
import           Data.List (unfoldr)
-- Ambiguous occurrence of chord/pitch names in haskore and HarmTrace
import qualified HarmTrace.Base.MusicRep                as MusicRep
import qualified HarmTrace.Harmony.ChordRep             as ChordRep  
import           HarmTrace.Harmony.Testing 
import           Data.Maybe                   
import qualified Data.EventList.Relative.TimeBody       as EventList  
import           Data.Ratio

type MidiArrange   = (InstrMap.ChannelTable MidiMusic.Instr, StdContext, MidiMusic.T) 
type StdContext    = Context.T NonNeg.Float Float (RhyMusic.Note MidiMusic.Drum MidiMusic.Instr)
type TimeSig       = (Double,Double)
type Bar           = (Int,Int)
type Duration      = Rational
type ChordList     = [MusicRep.Chord MusicRep.ScaleDegree]
type PropChordList = [(MusicRep.Chord MusicRep.ScaleDegree, Double)]
type MelodyTone    = (MusicRep.Root,Octave,Duration,ChordList)
data MelTone       = Int | Root 
                             deriving (Show,Eq)

-- data type song that is created from a midi file
data Song     = Song { key     :: MusicRep.Key
                     , timesig :: TimeSig 
                     , skip    :: Int
                     , melody  :: [(Bar,[MelodyTone])] } deriving(Eq)     
-- pretty show function for the song data type
instance Show Song where
        show (Song k t _ m) = "Key = " ++ (show k) ++ "\n" ++ "Time Signature = " ++ (show t) ++ "\n" ++ "Melody = " ++ (showMelody m) where
                showMelody (x:xs) = (show x) ++ "\n         " ++ showMelody xs
                showMelody []     = "" 

-- This function converts from a midi file the first channel to a tupel (HarmTrace Root, Octave (Int), Duration (NonNeg.T Rational))
-- It relies on information in the header of the midi file. If no key is found, a key of C maj is assumed. Likewise, if no TimeSig
-- is found, 4/4 is assumed. Because of the handling of the melodies, the midi file should not contain more than one track.   
-- Because of a bug in Sound.Midi, the key must be given seperately
midi2HarmTrace :: FilePath -> IO(Maybe Song)
midi2HarmTrace path = do midifile <- LoadMidi.fromFile path
                         contents <- fmap ReadMidi.toGMMusic (LoadMidi.fromFile path)
                         let chtable = fst3 (contents::MidiArrange) --InstrMap.ChannelTable MidiMusic.Instr
                         let context = snd3 (contents::MidiArrange) --StdContext
                         let music   = thd3 (contents::MidiArrange) --MidiMusic.T)
                         
                         -- check if the midi file is a single track file
                         if (length chtable /= 1)
                          then error ("midi2HarmTrace: Incomplete file header or more than one track in the midi file, found tracks: " ++ (show $ length chtable))
                           else let -- extract the melody from the midi file
                                    htflat = htMelodyFromMIDIFile music  
                                    -- fix the note lengths
                                    -- fixeddurs = map fixDuration htflat
                                    
                                    htKey = getKeyFromMIDIfile midifile
                                    (n,d) = getTimeSigFromMIDIfile midifile
                                    
                                    sep = seperateMelodyByBeat htflat (n,d) 
                                    songmel = createSongMelody sep (n,d) (1,fromIntegral(length(sep)))
                                    skip = calcSkip (n,d)
                                    song = Song htKey (n,d) skip songmel
                                in return (Just song)

seperateMelodyByBeat :: [(MusicRep.Root,Octave,Duration,ChordList)] -> TimeSig -> [[(MusicRep.Root,Octave,Duration,ChordList)]]
seperateMelodyByBeat melody beat = map reverse $ ph2 melody beat [] []    

-- Phrasing
ph2 :: [(MusicRep.Root,Octave,Duration,ChordList)] -> TimeSig -> [Rational] -> [(MusicRep.Root,Octave,Duration,ChordList)] -> [[(MusicRep.Root,Octave,Duration,ChordList)]]
ph2 ((note,o,dur,l):xs) (b,b2) a a2 | (sum(a) /= (truncate(b) % truncate(b2))) = ph2 xs (b,b2) (dur:a) ((note,o,dur,l):a2)
                                    | otherwise                                = a2:(ph2 ((note,o,dur,l):xs) (b,b2) [] [])
ph2 [] n a a2                       = [a2] 



{- This function gives us for the time signatures 1/$d$ a skip value of 1, signatures 3/$d$ and 6/$d$ a skip value of 3, for signatures 4/$d$ 
a skip value of 2 and for 8/$d$ a skip value of 4, where $d$ indicates the note value which represents one beat. The skip value tells us 
that every $s$ melody notes, there should be a chord, where $s$ is the skip value. -}
calcSkip :: TimeSig -> Int
calcSkip (n,_) | (n == 1) = 1
               | (n == 2) = 2 
               | (n == 3) = 3
               | (n == 4) = 2
               | (n == 6) = 3
               | (n == 8) = 4
      

-- guess the correct note length, to not have to deal with (241 % 1920) kind of notes    
-- not needed when using lilypond midi files
repairNotes' :: [MelodyTone] -> [MelodyTone]
repairNotes' ((r,o,d,l):xs)       | (d < 3.25) && (d > 2.75) = (r,o,(3%1),l):repairNotes' xs 
                                  | (d < 2.25) && (d > 1.75) = (r,o,(2%1),l):repairNotes' xs 
                                  | (d < 1.25) && (d > 0.75) = (r,o,(1%1),l):repairNotes' xs 
                                  | (d < 0.75) && (d > 0.25) = (r,o,(1%2),l):repairNotes' xs 
                                  | (d < 1%2) && (d > 1%6) = (r,o,(1%4),l):repairNotes' xs 
                                  | (d < 1%6) && (d > 1%10) = (r,o,(1%8),l):repairNotes' xs                                                       
                                  | (d < 1%14) && (d > 1%18) = (r,o,(1%16),l):repairNotes' xs                                                       
                                  | (d < 1%30) && (d > 1%34) = (r,o,(1%32),l):repairNotes' xs                                                       
                                  | (d < 1%62) && (d > 1%60) = (r,o,(1%64),l):repairNotes' xs
repairNotes' []           = []       

-- create a Song melody from the data from a midi file
createSongMelody :: [[(MusicRep.Root, Octave, Duration, ChordList)]] -> TimeSig -> Bar -> [(Bar,[MelodyTone])]
createSongMelody (m:ms) t (i,l) = ((i,l),m):createSongMelody (ms) t ((i+1),l)
createSongMelody [] _ _         = []

-- Divides all notes by 2. This is needed because of a bug in Sound.Midi
fixDuration :: (MusicRep.Root, Octave, Duration, ChordList) -> (MusicRep.Root, Octave, Duration,ChordList)
--fixDuration (r,o,d,l) = (r,o,(d/2),l)
fixDuration (r,o,d,l) = (r,o,d,l)

htMelodyFromMIDIFile :: (Eq t1) => CtrlMediumList.T t2 (Music.Primitive (RhyMusic.Note t t1)) -> [(MusicRep.Root, Octave, Duration,ChordList)]
htMelodyFromMIDIFile midifile = filterDeadNotes$ map midi2Root $ map fromJust $ filter (/= Nothing) $ map crtlMList2Note $ flattenCtrlList midifile

getKeyFromMIDIfile :: MidiFile.T -> MusicRep.Key
getKeyFromMIDIfile midifile =  metaKey2HTKey $ getKeyMeta metaEvents where 
                                        zeroTimeEvents = get0TimeEvents $ eventToList (head $ take 1 $ drop 1 $ getAllTracks midifile) -- (head $ getAllTracks midifile) -- When key is in header
                                        metaEvents = getMetaEvents zeroTimeEvents

getTimeSigFromMIDIfile :: MidiFile.T -> TimeSig
getTimeSigFromMIDIfile midifile = metaTimeSig2TimeSig $ getTimeSigMeta $ metaEvents where 
                                        zeroTimeEvents = get0TimeEvents $ eventToList (head $ getAllTracks midifile)
                                        metaEvents = getMetaEvents zeroTimeEvents
                        
metaKey2HTKey :: (KeySignature.Mode, KeySignature.Accidentals) -> MusicRep.Key 
metaKey2HTKey (k,s) = MusicRep.Key (metaKey2HTNote s) (metaScale2Mode k)  

-- if no time signature is found in the midi file, 4/4 is assumed
metaTimeSig2TimeSig :: (Int,Int,Int,Int) -> TimeSig
metaTimeSig2TimeSig (-1,-1,_,_) = (4,4)
metaTimeSig2TimeSig (a,b,_,_)   = (fromIntegral a, d) where 
                                      d = ((1 :: Double) / ((1::Double) / (2 :: Double) ^(b)))

-- transform scale quality midi meta information to HT representation
--metaScale2Mode :: MetaEvent.Scale -> MusicRep.Mode        
metaScale2Mode s | (s == KeySignature.Major) = MusicRep.MajMode 
                 | (s == KeySignature.Minor) = MusicRep.MinMode

-- transform scale key midi meta information to HT representation
metaKey2HTNote :: KeySignature.Accidentals -> MusicRep.Note MusicRep.DiatonicNatural
metaKey2HTNote k | (k == (KeySignature.Accidentals (-7))) = MusicRep.Note (Just MusicRep.Fl) MusicRep.C 
                 | (k == (KeySignature.Accidentals (-6))) = MusicRep.Note (Just MusicRep.Fl) MusicRep.G
                 | (k == (KeySignature.Accidentals (-5))) = MusicRep.Note (Just MusicRep.Fl) MusicRep.D
                 | (k == (KeySignature.Accidentals (-4))) = MusicRep.Note (Just MusicRep.Fl) MusicRep.A
                 | (k == (KeySignature.Accidentals (-3))) = MusicRep.Note (Just MusicRep.Fl) MusicRep.E
                 | (k == (KeySignature.Accidentals (-2))) = MusicRep.Note (Just MusicRep.Fl) MusicRep.B 
                 | (k == (KeySignature.Accidentals (-1))) = MusicRep.Note (Nothing) MusicRep.F
                 | (k == (KeySignature.Accidentals 0))  = MusicRep.Note (Nothing) MusicRep.C
                 | (k == (KeySignature.Accidentals 1))  = MusicRep.Note (Nothing) MusicRep.G
                 | (k == (KeySignature.Accidentals 2))  = MusicRep.Note (Nothing) MusicRep.D
                 | (k == (KeySignature.Accidentals 3))  = MusicRep.Note (Nothing) MusicRep.A
                 | (k == (KeySignature.Accidentals 4))  = MusicRep.Note (Nothing) MusicRep.E
                 | (k == (KeySignature.Accidentals 5))  = MusicRep.Note (Nothing) MusicRep.B
                 | (k == (KeySignature.Accidentals 6))  = MusicRep.Note (Just MusicRep.Sh) MusicRep.F
                 | (k == (KeySignature.Accidentals 7))  = MusicRep.Note (Just MusicRep.Sh) MusicRep.C
                 | otherwise                            = MusicRep.Note (Nothing) MusicRep.N

-- retrieve all 0 time midi events
get0TimeEvents :: [(ElapsedTime, Event.T)] -> [Event.T]
get0TimeEvents ((t,e):xs)  | (t == 0)  = e:get0TimeEvents xs
                           | otherwise = []
get0TimeEvents []          = []
 
-- retreive all midi meta events
getMetaEvents :: [Event.T] -> [MetaEvent.T]
getMetaEvents ((Event.MetaEvent x):xs) = x:getMetaEvents xs
getMetaEvents (o:os)                   = getMetaEvents os
getMetaEvents []                       = []

-- retrieve tracks from the midi file
getAllTracks :: MidiFile.T -> [Track]
getAllTracks (MidiFile.Cons _ _ trks) = trks

-- When no KeySig metadata is found in the midi header, a key of C major is assumed
getKeyMeta :: [MetaEvent.T] -> (KeySignature.Mode, KeySignature.Accidentals)
getKeyMeta ((MetaEvent.KeySig (KeySignature.Cons mode acc)):xs) = (mode,acc)
getKeyMeta (o:os)                  = getKeyMeta os
getKeyMeta []                      = (KeySignature.Major, (KeySignature.Accidentals 0))    

-- retrieve the time signature from the midi meta information
getTimeSigMeta :: [MetaEvent.T] -> (Int,Int,Int,Int)
getTimeSigMeta ((MetaEvent.TimeSig a b c d):xs) = (a,b,c,d)
getTimeSigMeta (o:os)                           = getTimeSigMeta os
getTimeSigMeta []                               = (-1,-1,-1,-1)    

-- transform midi track events to a tuple list
eventToList :: MidiFile.Track -> [(ElapsedTime, Event.T)]
eventToList a  | (a == EventList.empty) = []
               | otherwise              =  (x):eventToList y where 
                                                (x,y)  = fromJust $ EventList.viewL a

-- if no instruments are set in the midi file, assume AcousticGrandPiano                                              
fixTable :: InstrMap.ChannelTable MidiMusic.Instr -> InstrMap.ChannelTable MidiMusic.Instr
fixTable [] = [(GM.AcousticGrandPiano, MidiMusic.toChannel 0)]
fixTable o  = o
                      

midi2Root :: (Maybe Class,Octave,Duration,Rational,t1) -> (MusicRep.Root,Octave,Duration,ChordList)
midi2Root (clss,o,dur,velocity,instr) = (justClassToHT(clss),o,dur,[])

-- A note is considered dead when its duration is shorter than 1/64th    
-- This does nothing with a midi file generated from lilypond, as lilypond generated clean files
-- Other midi writers seem to include noisy short empty notes 
filterDeadNotes :: [(MusicRep.Root,Octave,Duration,ChordList)] -> [(MusicRep.Root,Octave,Duration,ChordList)]
filterDeadNotes ((r,o,d,l):xs) | (d > (1/64)) = (r,o,d,l): filterDeadNotes xs
                               | otherwise    = filterDeadNotes xs
filterDeadNotes []             = []

-- convert a Control Medium List to a 
crtlMList2Note :: CtrlMediumList.T t2 (Music.Primitive (RhyMusic.Note t t1)) -> Maybe (Maybe Class,Octave,Duration,Rational,Maybe t1)
crtlMList2Note (CtrlMediumList.Primitive (Music.Atom dur (Just (RhyMusic.Note velocity (RhyMusic.Tone instr (o,clss)))))) = Just (Just clss,o,NonNeg.toNumber dur,velocity,Just instr) 
crtlMList2Note (CtrlMediumList.Primitive (Music.Atom dur (Nothing))) = Just (Nothing,0,NonNeg.toNumber dur,0,Nothing) 
crtlMList2Note o = Nothing

getClassFromNote :: (Duration,RhyMusic.Note t t1) -> Class
getClassFromNote (velocity,(RhyMusic.Note dur (RhyMusic.Tone instr (o,clss)))) = clss

getDurFromNote :: (Duration,RhyMusic.Note t t1) -> Rational  
getDurFromNote (velocity,(RhyMusic.Note dur (RhyMusic.Tone instr (o,clss)))) = dur

getInstrFromNote :: (Duration,RhyMusic.Note t t1) -> t1
getInstrFromNote (velocity,(RhyMusic.Note dur (RhyMusic.Tone instr (o,clss)))) = instr

getOctaveFromNote :: (Duration,RhyMusic.Note t t1) -> Octave
getOctaveFromNote (velocity,(RhyMusic.Note dur (RhyMusic.Tone instr (o,clss)))) = o

getInfoFromNote :: (Duration,RhyMusic.Note t t1) -> (Class,Octave,Duration,Rational,t1)
getInfoFromNote (velocity,(RhyMusic.Note dur (RhyMusic.Tone instr (o,clss)))) = (clss,o,velocity,dur,instr)

-- Flatten the control medium list to obtain a flat list of substructures
-- eliminating parallel structures
flattenCtrlList :: CtrlMediumList.T t0 t1 -> [CtrlMediumList.T t0 t1]
flattenCtrlList  (CtrlMediumList.Serial   s) = flattenCtrlList' s
flattenCtrlList  (CtrlMediumList.Parallel s) = flattenCtrlList' s
flattenCtrlList' :: [CtrlMediumList.T t0 t1] -> [CtrlMediumList.T t0 t1]
flattenCtrlList' ((CtrlMediumList.Control x o):xs) = flattenCtrlList' xs
flattenCtrlList' ((CtrlMediumList.Primitive x):xs) = (CtrlMediumList.Primitive x): flattenCtrlList' xs
flattenCtrlList' [o] = flattenCtrlList o
flattenCtrlList' (x:xs) = flattenCtrlList' [x] ++ flattenCtrlList' xs
flattenCtrlList' []  = []                          

vline :: [NoteAttributes -> MelodyS.T] -> MelodyS.T
vline l = line (map ($ MelodyS.na) l)

v :: NoteAttributes
-- 
v = vel 1

vel :: Haskore.Basic.Dynamics.Velocity -> NoteAttributes
vel vl = Accessor.set MelodyS.velocity1 vl MelodyS.na

oct :: Octave
oct = 1

ch :: [Chord.Generic NoteAttributes] -> MelodyS.T          
ch c = line (map chord (Chord.leastVaryingInversions ((1,C),(1,C)) c))

extractMelody :: [(Bar,[MelodyTone])] -> [(MusicRep.Root,Octave,Duration)] 
extractMelody [] = []
extractMelody ((b,m):bs) = (extractMelody' m) ++ extractMelody bs where
        extractMelody' :: [MelodyTone] -> [(MusicRep.Root,Octave,Duration)]
        extractMelody' []             = [] 
        extractMelody' ((r,o,d,l):xs) = (r,o,d) : extractMelody' xs

-- Renders a [(MusicRep.Root,Double)] melody and a [MusicRep.Chord MusicRep.Root] harmonization into a midi file
--htToMidi :: String -> [(MusicRep.Root,Octave,Duration)] -> [MusicRep.Chord MusicRep.Root] -> IO ()
--htToMidi filename mel harm = render_to filename (htMelToHKMel mel) (map htChord2hkChord harm)
-- skip is the number of chords per bar, as given as an argument to the main function
toMidi :: TimeSig -> Int -> String -> [(Bar,[MelodyTone])] -> [MusicRep.Chord MusicRep.Root] -> IO ()
--toMidi t skip filename mel harm = render_to filename (htMelToHKMel $ extractMelody mel) (map (htChord2hkChord t skip) harm)    
toMidi t skip filename mel harm = render_to filename (htMelToHKMel $ extractMelody mel) (map (\(h,m) -> htChord2hkChord t skip h m) $ zip harm (extractMelody mel))

-- create a midi file from a melody and chords
-- the chords are placed an octave below the melody
render_to :: FilePath -> Melody.T () -> [Chord.Generic NoteAttributes] -> IO ()        
render_to f m c = Render.fileFromGeneralMIDIMusic f song where
                  song = MidiMusic.fromMelodyNullAttr MidiMusic.AcousticGrandPiano (loudness1 1.2 m) =:= 
                         MidiMusic.fromStdMelody MidiMusic.AcousticGrandPiano (loudness1 2 (transpose (-12) (ch c)))

-- htMelToHKMel converts a HarmTrace melody to a Haskore melody line
htMelToHKMel :: [(MusicRep.Root,Octave,Duration)] -> Music.T (Melody.Note ())
htMelToHKMel mel = line (map htNoteToHKNote mel) 

-- htNoteToHKNote converts HarmTrace Roots (Note DiatonicNatural) to Haskore Notes
-- 
htNoteToHKNote :: (MusicRep.Root,Octave,Duration) -> Melody.T ()
htNoteToHKNote (r,o,d) = (map (\e -> e o (NonNeg.fromNumber(d/2)) ()) [l])!!0 where
--htNoteToHKNote (r,o,d) = (map (\e -> e o (NonNeg.fromNumber(d)) ()) [l])!!0 where
                             l   = Melody.note' (rootToHK r)
                             --dur = htDurToHkDur d

-- htChord2hkChord converts a HarmTrace Chord Root to a Haskore Chord  
-- note lengths are doubled because of a bug in the Sound.Midi library                               
htChord2hkChord :: TimeSig -> Int -> MusicRep.Chord MusicRep.Root -> (MusicRep.Root,Octave,Duration) -> Chord.Generic NoteAttributes
htChord2hkChord (n,d) 0 (MusicRep.Chord chordRoot chordMode attr loc dur) (r,o,dm) = 
        Chord.generic (ChordRep.rootToClass(chordRoot)) (convertMode(chordMode)) (NonNeg.fromNumber $ (dm/2)) v
htChord2hkChord (n,d) skip (MusicRep.Chord chordRoot chordMode attr loc dur) (r,o,dm)  
        | (dm > (1%(floor d))*(toRational skip)) = Chord.generic (ChordRep.rootToClass(chordRoot)) (convertMode(chordMode)) (NonNeg.fromNumber $ (dm/2)) v
        | otherwise                              = Chord.generic (ChordRep.rootToClass(chordRoot)) (convertMode(chordMode)) (NonNeg.fromNumber $ (toRational((1%(floor d))*(toRational skip))/2)) v
--htChord2hkChord (n,d) skip (MusicRep.Chord chordRoot chordMode attr loc dur) _ = 
--        Chord.generic (ChordRep.rootToClass(chordRoot)) (convertMode(chordMode)) (NonNeg.fromNumber $ (toRational((1%(floor d))*(toRational skip))/2)) v

strongList :: TimeSig -> [Data.Ratio.Ratio Integer]
strongList (n,d) = take (floor n) $ [(floor i)%(floor d) | i <- [0..d-1]]

dr2 :: Int -> [a] -> [a]
dr2 m = unfoldr ((\x -> case listToMaybe x of
                         Nothing -> Nothing
                         Just i  -> Just (i,drop m x)))

-- note lengths are doubled because of a bug in the Sound.Midi library
htDurToHkDur :: Double -> Dur
htDurToHkDur dur = case dur of
      2.0      -> wn  -- brevis
      1.0      -> hn  -- whole note
      0.5      -> qn  -- half note      
      0.25     -> en
      0.125    -> sn
      0.0625   -> tn  -- sixteenth note
      0.03125  -> sfn  -- thirty-second note
      0.015625 -> sfn -- sixty-fourth note      

rootToHK :: MusicRep.Root -> Class
rootToHK root = case root of 
        MusicRep.Note (Just MusicRep.Fl) MusicRep.C -> Cf
        MusicRep.Note (Nothing) MusicRep.C          -> C
        MusicRep.Note (Just MusicRep.Sh) MusicRep.C -> Cs
        MusicRep.Note (Just MusicRep.Fl) MusicRep.D -> Df 
        MusicRep.Note (Nothing) MusicRep.D          -> D
        MusicRep.Note (Just MusicRep.Sh) MusicRep.D -> Ds 
        MusicRep.Note (Just MusicRep.Fl) MusicRep.E -> Ef
        MusicRep.Note (Nothing) MusicRep.E          -> E
        MusicRep.Note (Just MusicRep.Sh) MusicRep.E -> Es
        MusicRep.Note (Just MusicRep.Fl) MusicRep.F -> Ff 
        MusicRep.Note (Nothing) MusicRep.F          -> F 
        MusicRep.Note (Just MusicRep.Sh) MusicRep.F -> Fs
        MusicRep.Note (Just MusicRep.Fl) MusicRep.G -> Gf 
        MusicRep.Note (Nothing) MusicRep.G          -> G
        MusicRep.Note (Just MusicRep.Sh) MusicRep.G -> Gs 
        MusicRep.Note (Just MusicRep.Fl) MusicRep.A -> Af
        MusicRep.Note (Nothing) MusicRep.A          -> A
        MusicRep.Note (Just MusicRep.Sh) MusicRep.A -> As 
        MusicRep.Note (Just MusicRep.Fl) MusicRep.B -> Bf 
        MusicRep.Note (Nothing) MusicRep.B          -> B
        MusicRep.Note (Just MusicRep.Sh) MusicRep.B -> Bs
        
noteToHT :: Class -> MusicRep.Root
noteToHT root = case root of 
        Cf -> MusicRep.Note (Just MusicRep.Fl) MusicRep.C 
        C  -> MusicRep.Note (Nothing) MusicRep.C 
        Cs -> MusicRep.Note (Just MusicRep.Sh) MusicRep.C 
        Df -> MusicRep.Note (Just MusicRep.Fl) MusicRep.D 
        D  -> MusicRep.Note (Nothing) MusicRep.D 
        Ds -> MusicRep.Note (Just MusicRep.Sh) MusicRep.D 
        Ef -> MusicRep.Note (Just MusicRep.Fl) MusicRep.E 
        E  -> MusicRep.Note (Nothing) MusicRep.E 
        Es -> MusicRep.Note (Just MusicRep.Sh) MusicRep.E 
        Ff -> MusicRep.Note (Just MusicRep.Fl) MusicRep.F 
        F  -> MusicRep.Note (Nothing) MusicRep.F 
        Fs -> MusicRep.Note (Just MusicRep.Sh) MusicRep.F 
        Gf -> MusicRep.Note (Just MusicRep.Fl) MusicRep.G 
        G  -> MusicRep.Note (Nothing) MusicRep.G 
        Gs -> MusicRep.Note (Just MusicRep.Sh) MusicRep.G 
        Af -> MusicRep.Note (Just MusicRep.Fl) MusicRep.A 
        A  -> MusicRep.Note (Nothing) MusicRep.A 
        As -> MusicRep.Note (Just MusicRep.Sh) MusicRep.A 
        Bf -> MusicRep.Note (Just MusicRep.Fl) MusicRep.B 
        B  -> MusicRep.Note (Nothing) MusicRep.B 
        Bs -> MusicRep.Note (Just MusicRep.Sh) MusicRep.B 

justClassToHT :: Maybe Class -> MusicRep.Root
justClassToHT clss = case clss of 
        Just Cf -> MusicRep.Note (Just MusicRep.Fl) MusicRep.C 
        Just C  -> MusicRep.Note (Nothing) MusicRep.C 
        Just Cs -> MusicRep.Note (Just MusicRep.Sh) MusicRep.C 
        Just Df -> MusicRep.Note (Just MusicRep.Fl) MusicRep.D 
        Just D  -> MusicRep.Note (Nothing) MusicRep.D 
        Just Ds -> MusicRep.Note (Just MusicRep.Sh) MusicRep.D 
        Just Ef -> MusicRep.Note (Just MusicRep.Fl) MusicRep.E 
        Just E  -> MusicRep.Note (Nothing) MusicRep.E 
        Just Es -> MusicRep.Note (Just MusicRep.Sh) MusicRep.E 
        Just Ff -> MusicRep.Note (Just MusicRep.Fl) MusicRep.F 
        Just F  -> MusicRep.Note (Nothing) MusicRep.F 
        Just Fs -> MusicRep.Note (Just MusicRep.Sh) MusicRep.F 
        Just Gf -> MusicRep.Note (Just MusicRep.Fl) MusicRep.G 
        Just G  -> MusicRep.Note (Nothing) MusicRep.G 
        Just Gs -> MusicRep.Note (Just MusicRep.Sh) MusicRep.G 
        Just Af -> MusicRep.Note (Just MusicRep.Fl) MusicRep.A 
        Just A  -> MusicRep.Note (Nothing) MusicRep.A 
        Just As -> MusicRep.Note (Just MusicRep.Sh) MusicRep.A 
        Just Bf -> MusicRep.Note (Just MusicRep.Fl) MusicRep.B 
        Just B  -> MusicRep.Note (Nothing) MusicRep.B 
        Just Bs -> MusicRep.Note (Just MusicRep.Sh) MusicRep.B 
        Nothing -> MusicRep.Note (Nothing) MusicRep.N 

-- convertMode converts the HarmTrace Chord Shorthand to Haskore a Relative list
convertMode :: MusicRep.Shorthand -> [Relative]
convertMode mode = case mode of 
        MusicRep.Maj -> Chord.majorInt
        MusicRep.Min -> Chord.minorInt
        MusicRep.Dim -> [Interval.unison, Interval.minorThird, 6]
        --MusicRep.Aug
        -- Seventh chords
        MusicRep.Maj7 -> Chord.majorSeventhInt
        MusicRep.Min7 -> Chord.minorSeventhInt
        MusicRep.Sev  -> Chord.dominantSeventhInt
        --MusicRep.Dim7 
        --MusicRep.HDim7 
        MusicRep.MinMaj7 -> Chord.minorMajorSeventhInt
        -- Sixth chords
        --MusicRep.Maj6 
        --MusicRep.Min6
        -- Extended chords
        --MusicRep.Nin 
        --MusicRep.Maj9 
        --MusicRep.Min9
        -- Suspended chords
        MusicRep.Sus4  -> Chord.sustainedFourthInt
        --MusicRep.Sus2
        --MusicRep.None

-- retrieve the roots from the melody
getRoots :: [(MusicRep.Root,Double)] -> [MusicRep.Root]
getRoots ((r,d):xs) = r:getRoots xs
getRoots []         = []

-- retrieve the durations from the melody
getDoubles :: [(MusicRep.Root,Double)] -> [Double]
getDoubles ((r,d):xs) = d:getDoubles xs
getDoubles []         = []      

-- MIDI file to MidiMusic.T, a InstrMap, and a Context.
tcd :: FilePath -> IO ()
tcd file = do x <- fmap ReadMidi.toGMMusic (LoadMidi.fromFile file)
              putStrLn (show $ fst3 (x::MidiArrange))
              putStrLn (show $ snd3 x)
              putStrLn (show $ thd3 x)

-- A MIDI file to a MidiFile datatype and back to a MIDI file.
tcb :: FilePath -> IO ()
tcb file = LoadMidi.fromFile file >>= SaveMidi.toFile "test.mid"

--MIDI file to MidiFile datatype.
tc :: FilePath -> IO ()
tc file = LoadMidi.fromFile file >>= print

-- A MIDI file to MidiMusic.T and back to a MIDI file.
tcdab :: FilePath -> IO ()
tcdab file =
    LoadMidi.fromFile file >>=
       (SaveMidi.toFile "test.mid" . WriteMidi.fromGMMusic .
          (id::MidiArrange -> MidiArrange) . ReadMidi.toGMMusic)    

-- Number of bars
noBars :: [(Bar,[MelodyTone])] -> Int
noBars (((n,d),x):xs) = d